import pandas as pd
import numpy as np

def percent_analysis(clusters: dict,
                     threshold: float,
                     interactions: pd.DataFrame,
                     cluster_interactions: list,
                     base_result: pd.DataFrame,
                     separator: str,
                     ) -> pd.DataFrame:
    result = base_result.copy()
    percents = {}
    for cluster_name in clusters['names']:
        counts = clusters['counts'][cluster_name]

        percents[cluster_name] = counts.apply(lambda count: counts_percent(count, threshold), axis=1)

    for interaction_index, interaction in interactions.iterrows():
        for cluster_interaction in cluster_interactions:
            cluster_interaction_string = '{}{}{}'.format(cluster_interaction[0], separator, cluster_interaction[1])

            interaction_percent = cluster_interaction_percent(cluster_interaction, interaction, percents)

            result.at[interaction_index, cluster_interaction_string] = interaction_percent

    return result


def counts_percent(counts: pd.Series,
                   threshold: float) -> int:
    total = len(counts)
    positive = len(counts[counts > 0])

    if (positive / total) > threshold:
        return 1
    else:
        return 0


def cluster_interaction_percent(cluster_interaction: tuple,
                                interaction: pd.Series,
                                clusters_percents: dict,
                                ) -> int:
    percent_cluster_receptors = clusters_percents[cluster_interaction[0]]
    percent_cluster_ligands = clusters_percents[cluster_interaction[1]]

    receptor = interaction['multidata_1_id']
    percent_receptor = percent_cluster_receptors.loc[receptor]
    ligand = interaction['multidata_2_id']
    percent_ligand = percent_cluster_ligands.loc[ligand]

    if percent_receptor and percent_ligand:
        interaction_percent = 1

    else:
        interaction_percent = 0

    return interaction_percent


def get_significant_means(mean_analysis: pd.DataFrame,
                          result_percent: pd.DataFrame) -> pd.DataFrame:
    significant_means = mean_analysis.copy()
    for index, mean_analysis in mean_analysis.iterrows():
        for cluster_interaction in list(result_percent.columns):
            if not result_percent.at[index, cluster_interaction]:
                significant_means.at[index, cluster_interaction] = np.nan
    return significant_means


def build_significant_means(mean_analysis: pd.DataFrame,
                            result_percent: pd.DataFrame) -> (pd.Series, pd.DataFrame):
    """
    Calculates the significant means and add rank (number of non empty entries divided by total entries)
    """
    significant_means = get_significant_means(mean_analysis, result_percent)
    significant_mean_rank = significant_means.count(axis=1)  # type: pd.Series
    number_of_clusters = len(significant_means.columns)
    significant_mean_rank = significant_mean_rank.apply(lambda rank: rank / number_of_clusters)
    significant_mean_rank = significant_mean_rank.round(3)
    significant_mean_rank.name = 'rank'
    return significant_mean_rank, significant_means



import pandas as pd


def validate_genes(genes: pd.DataFrame) -> bool:
    are_valid = True

    duplicated_genes = genes[genes.duplicated(keep=False)]

    if not duplicated_genes.empty:
        print('SOME GENES ARE DUPLICATED:')
        print(duplicated_genes.sort_values('ensembl').to_csv(index=False))
        are_valid = False

    return are_valid



import pandas as pd
import numpy as np


def set_defaults(df: pd.DataFrame, defaults: dict, quiet=False) -> pd.DataFrame:
    df = df.copy()
    for column_name, default_value in defaults.items():
        if column_name not in df:
            if not quiet:
                print('missing column in dataframe: {}, set to default {}'.format(column_name, default_value))
            df[column_name] = default_value
            continue

        df[column_name].replace({np.nan: default_value}, inplace=True)

    return df



import pandas as pd

from cellphonedb.src.core.core_logger import core_logger


def filter_by_any_multidatas(multidatas: pd.DataFrame, interactions: pd.DataFrame) -> pd.DataFrame:
    """
    Filters interactions if any interactions multidatas are in multidatas list
    """
    interactions_filtered = pd.merge(multidatas, interactions, left_on='id_multidata', right_on='multidata_1_id')
    interactions_filtered = interactions_filtered.append(
        pd.merge(multidatas, interactions, left_on='id_multidata', right_on='multidata_2_id'))

    interactions_filtered.drop_duplicates('id_interaction', inplace=True)
    interactions_filtered.reset_index(drop=True, inplace=True)
    return interactions_filtered[interactions.columns.values]


def filter_by_multidatas(multidatas: pd.DataFrame, interactions: pd.DataFrame) -> pd.DataFrame:
    """
    Filters interactions if both interactions multidatas are in multidatas list
    """
    interactions_filtered = pd.merge(multidatas, interactions, left_on='id_multidata', right_on='multidata_1_id')
    interactions_filtered = pd.merge(multidatas, interactions_filtered, left_on='id_multidata',
                                     right_on='multidata_2_id')

    interactions_filtered.drop_duplicates('id_interaction', inplace=True)
    interactions_filtered.reset_index(drop=True, inplace=True)

    return interactions_filtered[interactions.columns.values]


def filter_by_min_score2(interactions: pd.DataFrame, min_score2: float):
    filtered_interactions = interactions[interactions['score_2'] > min_score2]

    return filtered_interactions


def filter_by_receptor_ligand_integrin(proteins: pd.DataFrame, interactions: pd.DataFrame) -> pd.DataFrame:
    """
    Returns a DataFrame of enabled integrin interactions
    """
    core_logger.debug('Filtering by integrin')
    multidata_receptors = proteins[proteins['integrin']]

    receptor_interactions = pd.merge(multidata_receptors, interactions, left_on='id_multidata',
                                     right_on='multidata_1_id')
    enabled_interactions = pd.merge(proteins, receptor_interactions, left_on='id_multidata',
                                    right_on='multidata_2_id', suffixes=['_ligands', '_receptors'])

    receptor_interactions_inverted = pd.merge(multidata_receptors, interactions, left_on='id_multidata',
                                              right_on='multidata_2_id')

    enabled_interactions_inverted = pd.merge(proteins, receptor_interactions_inverted, left_on='id_multidata',
                                             right_on='multidata_1_id', suffixes=['_ligands', '_receptors'])

    enabled_interactions = enabled_interactions.append(enabled_interactions_inverted).reset_index(drop=True)

    enabled_interactions.drop_duplicates(inplace=True)

    return enabled_interactions



from sqlalchemy import Column, Integer, String, ForeignKey, Boolean

from cellphonedb.src.core.database.sqlalchemy_models import Base


class Interaction(Base):
    __tablename__ = 'interaction_table'

    id_interaction = Column(Integer, nullable=False, primary_key=True)
    id_cp_interaction = Column(String, nullable=False, unique=True)

    multidata_1_id = Column(Integer, ForeignKey('multidata_table.id_multidata'), nullable=False)
    multidata_2_id = Column(Integer, ForeignKey('multidata_table.id_multidata'), nullable=False)

    source = Column(String)
    annotation_strategy = Column(String)



import itertools

import pandas as pd

from cellphonedb.src.core.core_logger import core_logger


def get_cluster_combinations(cluster_names):
    return list(itertools.product(cluster_names, repeat=2))


def apply_threshold(cluster_counts: pd.DataFrame, cluster_names: list, threshold: float) -> pd.DataFrame:
    """
    Sets to 0 minor value colunts than threshold
    """

    core_logger.debug('Aplicating Threshold {}'.format(threshold))
    cluster_counts_filtered = cluster_counts.copy()
    for cluster_name in cluster_names:
        cluster_counts_filtered.loc[
            cluster_counts_filtered[cluster_name] < float(threshold), [cluster_name]] = 0.0

    return cluster_counts_filtered


def expand_multidata(cluster_counts: pd.DataFrame, multidatas: pd.DataFrame) -> pd.DataFrame:
    expanded = pd.merge(cluster_counts, multidatas, left_on='id_multidata', right_on='id_multidata')
    expanded = expanded.drop_duplicates().reset_index(drop=True)

    return expanded



import pandas as pd


def only_noncomplex_interactions(interactions, complexes):
    """

    :type interactions: pd.DataFrame
    :type complexes: pd.DataFrame
    :rtype: pd.DataFrame
    """

    proteins_in_complex = []

    for i in range(1, 5):
        proteins_in_complex = proteins_in_complex + complexes['uniprot_%s' % i].dropna().tolist()

    proteins_in_complex = list(set(proteins_in_complex))

    inweb_df_no_complex = interactions[interactions['uniprot_1'].apply(
        lambda protein: protein not in proteins_in_complex
    )]
    inweb_df_no_complex = inweb_df_no_complex[
        inweb_df_no_complex['uniprot_2'].apply(
            lambda protein: protein not in proteins_in_complex
        )]

    return inweb_df_no_complex



import pandas as pd

from cellphonedb.src.core.utils import filters


def call(proteins: pd.DataFrame, multidata_colums, protein_columns):
    bools = ['transmembrane', 'secreted', 'peripheral', 'receptor', 'other', 'secreted_highlight']

    proteins.rename(index=str, columns={'uniprot': 'name'}, inplace=True)
    proteins[bools] = proteins[bools].astype(bool)

    proteins = _optimizations(proteins)
    multidatas_to_add = process_multidata_data(proteins, multidata_colums)
    proteins_to_add = process_protein_data(proteins, protein_columns)

    return proteins_to_add, multidatas_to_add


def _optimizations(proteins):
    proteins['is_complex'] = False
    return proteins


def process_multidata_data(proteins: pd.DataFrame, multidata_columns: list) -> pd.DataFrame:
    multidata_proteins = proteins.copy()
    return filters.remove_not_defined_columns(multidata_proteins, multidata_columns)


def process_protein_data(proteins: pd.DataFrame, protein_columns: list) -> pd.DataFrame:
    proteins = proteins.copy()
    proteins = filters.remove_not_defined_columns(proteins, protein_columns + ['name'])
    return proteins



import logging

import sys

app_logger = logging.getLogger(__name__)


def setLevel(level: str = 'WARNING'):
    app_logger.setLevel(getattr(logging, level))


if not app_logger.handlers:
    app_logger.propagate = False
    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('[ ][APP][%(asctime)s][%(levelname)s] %(message)s', "%d/%m/%y-%H:%M:%S")
    ch.setFormatter(formatter)

    setLevel()
    app_logger.addHandler(ch)



import pandas as pd


def call(proteins_expanded: pd.DataFrame) -> pd.DataFrame:
    proteins_expanded.rename(index=str, columns={'name': 'uniprot'}, inplace=True)
    return proteins_expanded[
        ['uniprot', 'protein_name', 'transmembrane', 'peripheral', 'secreted', 'secreted_desc', 'secreted_highlight',
         'receptor', 'receptor_desc', 'integrin', 'other', 'other_desc', 'tags', 'tags_description', 'tags_reason',
         # 'pfam'
         ]].sort_values('uniprot')



class AllCountsFilteredException(Exception):
    def __init__(self, description: str = None, hint: str = None):
        super(AllCountsFilteredException, self).__init__('All counts filtered')
        self.description = description
        self.hint = hint



from flask_testing import TestCase

from cellphonedb.src.app.flask.flask_app import create_app


class TestDatabase(TestCase):
    def create_app(self):
        return create_app(environment='test', raise_non_defined_vars=False)

    def test_database_init(self):
        pass



import sys
import traceback

import click

from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.exceptions.MissingR import MissingR
from cellphonedb.src.exceptions.RRuntimeException import RRuntimeException
from cellphonedb.src.plotters import r_plotter

try:
    r_plotter.ensure_R_setup()
except MissingR:
    print('As there is no R environment set up, some functionalities will be disabled, e.g. plot')


@click.command()
@click.option('--means-path', type=click.Path(exists=True, file_okay=True, dir_okay=False),
              default='./out/means.txt', help='Analysis output means [./out/means.txt]')
@click.option('--pvalues-path', type=click.Path(exists=True, file_okay=True, dir_okay=False),
              default='./out/pvalues.txt', help='Analysis output pvalues [./out/pvalues.txt]')
@click.option('--output-path', type=click.Path(file_okay=False, writable=True),
              default='./out', help='Path to write generated plot [./out]')
@click.option('--output-name', type=str, default='./plot.pdf', help='Output file with plot [plot.pdf]')
@click.option('--rows', type=click.Path(exists=True, file_okay=True, dir_okay=False),
              help='List of rows to plot, one per line [all available]')
@click.option('--columns', type=click.Path(exists=True, file_okay=True, dir_okay=False),
              help='List of columns to plot, one per line [all available]')
@click.option('--verbose/--quiet', default=True, help='Print or hide cellphonedb logs [verbose]')
def dot_plot(means_path: str, pvalues_path: str, output_path: str, output_name: str, rows: str,
             columns: str, verbose: bool):
    try:
        r_plotter.dot_plot(means_path=means_path,
                           pvalues_path=pvalues_path,
                           output_path=output_path,
                           output_name=output_name,
                           rows=rows,
                           columns=columns)
    except MissingR:
        print('You cannot perform this plot command unless there is a working R setup according to CellPhoneDB specs')
    except RRuntimeException as e:
        app_logger.error(str(e))
    except:
        app_logger.error('Unexpected error')

        if verbose:
            traceback.print_exc(file=sys.stdout)
        else:
            app_logger.error('execute with --verbose to see full stack trace')


@click.command()
@click.argument('meta-path', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('--pvalues-path', type=click.Path(exists=True, file_okay=True, dir_okay=False),
              default='./out/pvalues.txt', help='Analysis output pvalues [./out/pvalues.txt]')
@click.option('--output-path', type=click.Path(file_okay=False, writable=True),
              default='./out', help='Path to write generated plot [./out]')
@click.option('--count-name', type=str, default='heatmap_count.pdf',
              help='Output file with count plot [heatmap_count.pdf]')
@click.option('--log-name', type=str, default='heatmap_log_count.pdf',
              help='Output file with log count plot [heatmap_log_count.pdf]')
@click.option('--count-network-name', type=str, default='count_network.txt',
              help='Output file name for count network [count_network.txt]')
@click.option('--interaction-count-name', type=str, default='interaction_count.txt',
              help='Output file name for interaction count [interaction_count.txt]')
@click.option('--pvalue', type=float, default=0.05, help='pvalue threshold to consider when plotting [0.05]')
@click.option('--verbose/--quiet', default=True, help='Print or hide cellphonedb logs [verbose]')
def heatmap_plot(meta_path: str, pvalues_path: str, output_path: str, count_name: str, log_name: str,
                 count_network_name, interaction_count_name, pvalue: float, verbose: bool):
    try:
        r_plotter.heatmaps_plot(meta_file=meta_path,
                                pvalues_file=pvalues_path,
                                output_path=output_path,
                                count_name=count_name,
                                log_name=log_name,
                                count_network_filename=count_network_name,
                                interaction_count_filename=interaction_count_name,
                                pvalue=pvalue)
    except MissingR:
        print('You cannot perform this plot command unless there is a working R setup according to CellPhoneDB specs')
    except RRuntimeException as e:
        app_logger.error(str(e))
    except:
        app_logger.error('Unexpected error')

        if verbose:
            traceback.print_exc(file=sys.stdout)
        else:
            app_logger.error('execute with --verbose to see full stack trace')



import pandas as pd

from cellphonedb.tools.tools_helper import interaction_exist


def remove_interactions_in_file(interactions, interactions_to_remove):
    '''

    :type interactions: pd.DataFrame
    :type interactions_to_remove: pd.DataFrame
    :rtype: pd.DataFrame
    '''

    interactions_filtered = interactions[
        interactions.apply(lambda row: interaction_exist(row, interactions_to_remove), axis=1) == False]

    return interactions_filtered



from cellphonedb.src.app.app_logger import app_logger


class LocalQueryLauncher:
    def __getattribute__(self, name):
        method = object.__getattribute__(self, name)
        if hasattr(method, '__call__'):
            app_logger.info('Launching Query {}'.format(name))

        return method

    def __init__(self, cellphonedb_app):

        self.cellphonedb_app = cellphonedb_app

    def find_interactions_by_element(self, element: str) -> None:
        print(self.cellphonedb_app.query.find_interactions_by_element(element).to_csv(index=False))

    def get_interaction_gene(self, columns: str) -> None:
        if columns:
            columns = columns.split(',')

        print(self.cellphonedb_app.query.get_all_genes(columns).to_csv(index=False))

    def autocomplete_element(self, partial_element: str) -> None:
        print(self.cellphonedb_app.query.autocomplete_launcher(partial_element).to_csv(index=False))



import json

from flask import request, Response

from cellphonedb.src.app.cellphonedb_app import cellphonedb_app
from cellphonedb.src.api_endpoints.web_api.web_api_endpoint_base import WebApiEndpointBase


class WebEndpointQueryComplexDeconvoluted(WebApiEndpointBase):
    def post(self):
        parameters = json.loads(request.get_data(as_text=True))

        complex_name = parameters['complex_name']

        deconvoluted = cellphonedb_app.cellphonedb.query.get_complex_deconvoluted(complex_name)

        if deconvoluted.empty:
            self.attach_error(
                {'code': 'element_not_found', 'title': '%s is not CellPhoneDB Complex' % complex_name,
                 'details': '%s is not present in CellPhoneDB complex table' % complex_name})
        else:
            self._attach_csv(deconvoluted.to_csv(index=False, sep=','), 'result')

        self._commit_attachments()

        return Response(self._msg.as_string(), mimetype='multipart/form-data; boundary="%s"' % self._msg.get_boundary())



from cellphonedb.src.api_endpoints.web_api.query_web_api_endpoints.complex import query_complex_web_api_routes
from cellphonedb.src.api_endpoints.web_api.query_web_api_endpoints.interaction import query_interaction_web_api_routes
from cellphonedb.src.api_endpoints.web_api.query_web_api_endpoints.web_endpoint_query_autocomplete import \
    WebEndpointQueryAutocomplete
from cellphonedb.src.api_endpoints.web_api.query_web_api_endpoints.web_endpoint_query_database import \
    WebEndpointQueryDatabase
from cellphonedb.src.api_endpoints.web_api.query_web_api_endpoints.web_endpoint_query_find_interactions_by_element import \
    WebEndpointQueryFindInteractionsByElement


def add(api, prefix=''):
    query_interaction_web_api_routes.add(api, '{}/interaction'.format(prefix))
    query_complex_web_api_routes.add(api, '{}/complex'.format(prefix))

    api.add_resource(WebEndpointQueryFindInteractionsByElement, '{}/find_interactions_by_element'.format(prefix))
    api.add_resource(WebEndpointQueryAutocomplete, '{}/autocomplete'.format(prefix))
    api.add_resource(WebEndpointQueryDatabase, '{}/database_versions'.format(prefix))



import pandas as pd

from cellphonedb.src.core.models.interaction import interaction_properties


def filter_by_cellphonedb_interactor(uniprots: pd.DataFrame, interactions: pd.DataFrame) -> pd.DataFrame:
    interactions_filtered = pd.merge(interactions, uniprots, left_on=['protein_1'], right_on=['uniprot'], how='inner')

    interactions_filtered = pd.merge(interactions_filtered, uniprots, left_on=['protein_2'],
                                     right_on=['uniprot'], how='inner', suffixes=('_1', '_2'))

    interactions_filtered.rename(columns={'uniprot_1': 'id_multidata_1', 'uniprot_2': 'id_multidata_2'}, inplace=True)
    interactions_filtered.drop_duplicates(inplace=True)

    interactions_filtered = interactions_filtered[
        interactions_filtered.apply(lambda interaction: interaction_properties.is_cellphonedb_interactor(interaction),
                                    axis=1)]

    return interactions_filtered[interactions.columns.values]



import pandas as pd


def expand_interactions_multidatas(interactions: pd.DataFrame, multidatas_expanded,
                                   suffixes: list = ('_1', '_2')) -> pd.DataFrame:
    interactions_expanded = pd.merge(interactions, multidatas_expanded, left_on='multidata_1_id',
                                     right_on='id_multidata')

    interactions_expanded = pd.merge(interactions_expanded, multidatas_expanded, left_on='multidata_2_id',
                                     right_on='id_multidata', suffixes=suffixes)

    interactions_expanded.drop_duplicates(inplace=True)
    return interactions_expanded


def get_duplicated(interactions: pd.DataFrame, column_1: str, column_2: str) -> pd.DataFrame:
    def duplicated(row):
        if len(interactions[(row[column_1] == interactions[column_1]) & (
                row[column_2] == interactions[column_2])]) > 1:
            return True

        if (row[column_1] != row[column_2]) & len(
                interactions[(row[column_1] == interactions[column_2]) & (
                        row[column_2] == interactions[column_1])]) > 0:
            return True
        return False

    result = interactions[interactions.apply(duplicated, axis=1)]

    return result



from cellphonedb.src.api_endpoints.web_api.query_web_api_endpoints.interaction.web_endpoint_query_interaction_gene import \
    WebEndpointQueryInteractionGene


def add(api, prefix=''):
    api.add_resource(WebEndpointQueryInteractionGene, '{}/gene'.format(prefix))



import os

from cellphonedb.src.core.collectors.collector import Collector
from cellphonedb.src.core.exporters.exporterlauncher import ExporterLauncher
from cellphonedb.src.core.methods.method_launcher import MethodLauncher
from cellphonedb.src.core.database import DatabaseManager
from cellphonedb.src.core.queries.query_launcher import QueryLauncher

cellphone_core_dir = os.path.dirname(os.path.realpath(__file__))
data_dir = '{}/data'.format(cellphone_core_dir)

data_test_dir = '{}/tests/fixtures'.format(cellphone_core_dir)


class Cellphonedb(object):
    def __init__(self, database_manager: DatabaseManager, config):
        self.config = {'default_threads': config['threads']}
        self.database_manager = database_manager
        self.export = ExporterLauncher(self.database_manager)
        self.collect = Collector(self.database_manager)
        self.method = MethodLauncher(self.database_manager, self.config['default_threads'])
        self.query = QueryLauncher(self.database_manager)
        self.debug_mode = config['debug']



import os
from datetime import datetime
from typing import Optional

import click
from click import Context

from cellphonedb.src.api_endpoints.terminal_api.tools_terminal_api_endpoints.tools_terminal_commands import \
    generate_proteins, generate_complex, _set_paths, generate_interactions, filter_all, generate_genes
from cellphonedb.src.app.cellphonedb_app import output_dir
from cellphonedb.src.database.manager import DatabaseVersionManager
from cellphonedb.src.database.manager.DatabaseVersionManager import collect_database
from cellphonedb.utils.utils import set_paths


@click.command()
@click.option('--database', default='cellphone_custom_{}.db'.format(datetime.now().strftime("%Y-%m-%d-%H_%M")),
              help='output file name [cellphone_custom_<current date_time>.db]')
@click.option('--result-path', default='', help='output folder for the collected database')
def collect(database, result_path):
    output_path = set_paths(output_dir, result_path)

    DatabaseVersionManager.collect_database(database, output_path)


@click.command()
@click.option('--version', type=str, default='latest')
def download(version: str):
    DatabaseVersionManager.download_database(version)


@click.command()
def list_remote():
    DatabaseVersionManager.list_remote_database_versions()


@click.command()
def list_local():
    DatabaseVersionManager.list_local_database_versions()


@click.command()
@click.option('--user-protein', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.option('--user-gene', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.option('--user-complex', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.option('--user-interactions', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.option('--user-interactions-only', is_flag=True)
@click.option('--fetch', is_flag=True)
@click.option('--result-path', type=str, default=None)
@click.option('--log-file', type=str, default='log.txt')
@click.option('--project-name', type=str, default=None)
@click.pass_context
def generate(ctx: Context,
             user_protein: Optional[str],
             user_gene: Optional[str],
             user_complex: Optional[str],
             user_interactions: Optional[str],
             user_interactions_only: Optional[str],
             fetch: bool,
             result_path: Optional[str],
             log_file: str,
             project_name: str
             ):
    ctx.invoke(generate_proteins,
               user_protein=user_protein,
               fetch_uniprot=fetch,
               result_path=result_path,
               log_file=log_file,
               project_name=project_name
               )

    ctx.invoke(generate_genes,
               user_gene=user_gene,
               fetch_uniprot=fetch,
               fetch_ensembl=fetch,
               result_path=result_path,
               project_name=project_name
               )

    ctx.invoke(generate_complex,
               user_complex=user_complex,
               result_path=result_path,
               log_file=log_file,
               project_name=project_name
               )

    output_path = _set_paths(result_path, project_name)

    proteins_file = os.path.join(output_path, 'protein_generated.csv')
    genes_file = os.path.join(output_path, 'gene_generated.csv')
    complex_file = os.path.join(output_path, 'complex_generated.csv')

    ctx.invoke(generate_interactions,
               proteins=proteins_file,
               genes=genes_file,
               complex=complex_file,
               user_interactions=user_interactions,
               user_interactions_only=user_interactions_only,
               result_path=result_path,
               fetch_imex=fetch,
               fetch_iuphar=fetch,
               project_name=project_name
               )

    ctx.invoke(filter_all, input_path=output_path, result_path=result_path)

    db_name = 'cellphonedb_user_{}.db'.format(datetime.now().strftime("%Y-%m-%d-%H_%M"))

    collect_database(db_name, output_path,
                     protein_filename='protein_input.csv',
                     gene_filename='gene_input.csv',
                     complex_filename='complex_input.csv',
                     interaction_filename='interaction_input.csv',
                     data_path=output_path)


@click.command()
@click.argument('path', type=str)
@click.option('--result-path', type=str, default=None)
@click.option('--project-name', type=str, default=None)
def collect_generated(path: str, result_path: Optional[str], project_name: str):
    db_name = 'cellphonedb_user_{}.db'.format(datetime.now().strftime("%Y-%m-%d-%H_%M"))
    output_path = _set_paths(result_path, project_name)

    collect_database(db_name,
                     output_path,
                     protein_filename='{}/protein_input.csv'.format(path),
                     gene_filename='{}/gene_input.csv'.format(path),
                     complex_filename='{}/complex_input.csv'.format(path),
                     interaction_filename='{}/interaction_input.csv'.format(path),
                     data_path=output_path)



import pandas as pd

from cellphonedb.src.core.core_logger import core_logger
from cellphonedb.src.core.database import DatabaseManager
from cellphonedb.src.core.queries.autocomplete_queries import autocomplete_query
from cellphonedb.src.core.queries.complex import complex_deconvoluted
from cellphonedb.src.core.queries.interaction import interactions_by_element


class QueryLauncher:
    def __init__(self, database_manager: DatabaseManager):
        self.database_manager = database_manager

    def __getattribute__(self, name: str):
        method = object.__getattribute__(self, name)
        if hasattr(method, '__call__'):
            core_logger.info('Launching Query {}'.format(name))

        return method

    def autocomplete_launcher(self, partial_element) -> pd.DataFrame:
        multidatas = self.database_manager.get_repository('multidata').get_all()
        genes = self.database_manager.get_repository('gene').get_all_expanded()[
            ['name', 'ensembl', 'protein_name', 'gene_name', 'hgnc_symbol']]

        return autocomplete_query(genes, multidatas, partial_element)

    def find_interactions_by_element(self, element: str) -> pd.DataFrame:
        interactions = self.database_manager.get_repository('interaction').get_all_expanded(suffixes=('_a', '_b'))
        complexes_composition = self.database_manager.get_repository('complex').get_all_compositions_expanded()

        return interactions_by_element.call(element, interactions, complexes_composition)

    def get_all_genes(self, columns: list = None) -> list:
        default_columns = ['ensembl', 'hgnc_symbol', 'gene_name']
        genes = self.database_manager.get_repository('gene').get_all()[columns or default_columns]
        return genes

    def get_complex_deconvoluted(self, complex_name: str) -> pd.DataFrame:
        complex_compositions = self.database_manager.get_repository('complex').get_all_compositions_expanded()

        complex_elements = complex_deconvoluted.call(complex_compositions, complex_name)
        return complex_elements



from sqlalchemy import Column, Integer, String, ForeignKey

from cellphonedb.src.core.database.sqlalchemy_models import Base


class Complex(Base):
    __tablename__ = 'complex_table'
    id_complex = Column(Integer, nullable=False, primary_key=True)

    complex_multidata_id = Column(Integer, ForeignKey('multidata_table.id_multidata'), nullable=False, unique=True)
    pdb_structure = Column(String)
    pdb_id = Column(String)
    stoichiometry = Column(String)
    comments_complex = Column(String)



import os
import unittest

from cellphonedb.src.app.cellphonedb_app import output_test_dir, data_test_dir
from cellphonedb.src.exceptions.MissingR import MissingR
from cellphonedb.src.plotters.r_plotter import dot_plot, heatmaps_plot, ensure_R_setup


def r_available():
    try:
        ensure_R_setup()
        return True
    except MissingR:
        return False


@unittest.skipUnless(r_available(), 'R setup is not available in this setup')
class TestPlots(unittest.TestCase):
    def test_dot_plot_with_all_columns_and_rows(self):
        means_path = self._get_input_file_path_for('means')
        pvalues_path = self._get_input_file_path_for('pvalues')
        plot_output = 'dot_plot_output_all_rows_columns.png'

        dot_plot(means_path=means_path,
                 pvalues_path=pvalues_path,
                 output_path=output_test_dir,
                 output_name=plot_output
                 )

        output_path = os.path.join(output_test_dir, plot_output)
        self.assertTrue(os.path.exists(output_path), 'Plot of type dot_plot did not work')
        self.assertGreater(os.path.getsize(output_path), 0)
        os.remove(output_path)

    def test_dot_plot_with_custom_columns_and_rows(self):
        means_path = self._get_input_file_path_for('means')
        pvalues_path = self._get_input_file_path_for('pvalues')
        columns_file = self._get_fixture('dot_plot_columns.txt')
        rows_file = self._get_fixture('dot_plot_rows.txt')
        plot_output = 'dot_plot_output_custom_rows_columns.png'

        dot_plot(means_path=means_path,
                 pvalues_path=pvalues_path,
                 output_path=output_test_dir,
                 output_name=plot_output,
                 rows=rows_file,
                 columns=columns_file
                 )

        output_path = os.path.join(output_test_dir, plot_output)
        self.assertTrue(os.path.exists(output_path), 'Plot of type dot_plot did not work')
        self.assertGreater(os.path.getsize(output_path), 0)
        os.remove(output_path)

    def test_heatmaps_plot(self):
        meta_path = self._get_fixture('hi_test_meta.txt')
        pvalues_path = self._get_input_file_path_for('pvalues')
        plot_output_counts = 'heatmaps_plot_counts.png'
        plot_output_count_logs = 'heatmaps_plot_count_logs.png'
        count_network_filename = 'count_network.txt'
        interaction_count_filename = 'interaction_count.txt'


        heatmaps_plot(meta_file=meta_path,
                      pvalues_file=pvalues_path,
                      output_path=output_test_dir,
                      count_name=plot_output_counts,
                      log_name=plot_output_count_logs,
                      count_network_filename=count_network_filename,
                      interaction_count_filename=interaction_count_filename,
                      pvalue=0.05
                      )

        for file in [plot_output_counts, plot_output_count_logs]:
            output_path = os.path.join(output_test_dir, file)
            self.assertTrue(os.path.exists(output_path), 'Plot of type heatmap_plot did not work')
            self.assertGreater(os.path.getsize(output_path), 0)
            os.remove(output_path)

    @staticmethod
    def _get_input_file_path_for(kind):
        template = 'input_for_plot_statistical_analysis__{}_result__data-test_it-10_seed-0_threshold-01_precision-1.txt'

        return os.path.join(data_test_dir, template.format(kind))

    @staticmethod
    def _get_fixture(name):
        return os.path.join(data_test_dir, name)



from cellphonedb.src.app.cellphonedb_app import cellphonedb_app
from cellphonedb.src.app.flask.flask_app import create_app
from cellphonedb.src.tests.cellphone_flask_test_case import CellphoneFlaskTestCase


class TestValidatorDatabaseNumberOfEntries(CellphoneFlaskTestCase):
    def create_app(self):
        return create_app(raise_non_defined_vars=False, verbose=False)

    def test_protein(self):
        proteins = cellphonedb_app.cellphonedb.database_manager.get_repository('protein').get_all()

        self.assertEqual(978, len(proteins), 'Number of Protein entries are different')

    def test_gene(self):
        genes = cellphonedb_app.cellphonedb.database_manager.get_repository('gene').get_all()
        self.assertEqual(1252, len(genes), 'Number of Gene entries are different')

    def test_complex(self):
        complex = cellphonedb_app.cellphonedb.database_manager.get_repository('complex').get_all()
        self.assertEqual(112, len(complex), 'Number of Complex entries are different')

    def test_multidata(self):
        multidatas = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'multidata').get_all()
        self.assertEqual(1090, len(multidatas), 'Number of Multidata entries are different')

    def test_protein_complex(self):
        multidatas = cellphonedb_app.cellphonedb.database_manager.get_repository('multidata').get_all()
        number_of_multidata = len(multidatas)

        proteins = cellphonedb_app.cellphonedb.database_manager.get_repository('protein').get_all()
        number_of_protein = len(proteins)

        complexes = cellphonedb_app.cellphonedb.database_manager.get_repository('complex').get_all()
        number_of_complex = len(complexes)

        self.assertEqual(number_of_complex + number_of_protein, number_of_multidata,
                         'Number of multidata is diferent than proteins+complex')

    def test_interaction(self):
        interactions = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'interaction').get_all()

        self.assertEqual(1396, len(interactions), 'Number of interactions are not equal')

    def test_interaction_curated(self):
        interactions = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'interaction').get_all()
        self.assertEqual(878, len(interactions[interactions['annotation_strategy'] == 'curated']),
                         'Number of curated interactions not equal')

    def test_interaction_guidetopharmacology(self):
        interactions = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'interaction').get_all()
        self.assertEqual(156, len(interactions[interactions['annotation_strategy'] == 'guidetopharmacology.org']),
                         'Number of source=guidetopharmacology.org (iphar) entries is different')

    def test_complex_composition(self):
        complex_compositions = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'complex').get_all_compositions()
        self.assertEqual(230, len(complex_compositions), 'Number of Complex Composition entries are different')

    def setUp(self):
        return create_app(raise_non_defined_vars=False, verbose=False)



import pandas as pd

from cellphonedb.tools.tools_helper import normalize_interactions


def call(iuphar_guidepharmacology: pd.DataFrame, genes: pd.DataFrame, proteins: pd.DataFrame) -> pd.DataFrame:
    iuphar_filtered = iuphar_guidepharmacology[iuphar_guidepharmacology['target_species'] == 'Human']

    iuphar_filtered = iuphar_filtered[iuphar_filtered['ligand_species'] == 'Human']

    iuphar_filtered = _process_target_uniprot(iuphar_filtered, proteins)

    iuphar_filtered = _process_ligand_gene_symbol(genes, iuphar_filtered)
    iuphar_filtered_cellphone_format = pd.DataFrame()
    iuphar_filtered_cellphone_format[['uniprot_1', 'uniprot_2']] = iuphar_filtered[
        ['uniprot', 'target_uniprot']]

    iuphar_filtered_cellphone_format['annotation_strategy'] = 'guidetopharmacology.org'
    iuphar_filtered_cellphone_format['iuphar'] = True

    iuphar_procesed = _drop_duplicates(iuphar_filtered_cellphone_format)

    return iuphar_procesed


def _drop_duplicates(iuphar_filtered_cellphone_format: pd.DataFrame) -> pd.DataFrame:
    interactions_normalized = normalize_interactions(iuphar_filtered_cellphone_format, 'uniprot_1',
                                                     'uniprot_2')
    interactions_duplicated = interactions_normalized[
        interactions_normalized.duplicated(['uniprot_1', 'uniprot_2'])]
    # if not interactions_duplicated.empty:
    #     print('WARNING: SOME IUPHAR INTERACTIONS ARE DUPLICATED')
    #     print(interactions_duplicated.sort_values(['uniprot_1', 'uniprot_2']).to_csv(index=False))
    iuphar_procesed = interactions_normalized.drop_duplicates(['uniprot_1', 'uniprot_2'])
    return iuphar_procesed


def _process_ligand_gene_symbol(genes: pd.DataFrame, iuphar_filtered: pd.DataFrame) -> pd.DataFrame:
    iuphar_filtered.dropna(subset=['ligand_gene_symbol'], inplace=True)
    iuphar_filtered = iuphar_filtered[iuphar_filtered['ligand_gene_symbol'].apply(lambda uniprot: '|' not in uniprot)]
    missing_genes = iuphar_filtered[~iuphar_filtered['ligand_gene_symbol'].isin(genes['gene_name'])][
        'ligand_gene_symbol'].drop_duplicates()
    if not missing_genes.empty:
        print('WARNING: {} GENE NAMES IN GUIDETOPHARMACOLOGY DATABASE DIDNT EXIST IN CELLPHONEDB DATABASE'.format(
            len(missing_genes)))
        print(missing_genes.tolist())
    iuphar_filtered = pd.merge(iuphar_filtered, genes, left_on='ligand_gene_symbol', right_on='gene_name')
    return iuphar_filtered


def _process_target_uniprot(iuphar_filtered: pd.DataFrame, proteins: pd.DataFrame) -> pd.DataFrame:
    iuphar_filtered.dropna(subset=['target_uniprot'], inplace=True)
    iuphar_filtered = iuphar_filtered[iuphar_filtered['target_uniprot'].apply(lambda uniprot: '|' not in uniprot)]
    missing_uniprots = iuphar_filtered[~iuphar_filtered['target_uniprot'].isin(proteins['uniprot'])][
        'target_uniprot'].drop_duplicates()
    if not missing_uniprots.empty:
        print('WARNING: {} UNIPROTS IN GUIDETOPHARMACOLOGY DATABASE DIDNT EXIST IN CELLPHONEDB DATABASE'.format(
            len(missing_uniprots)))
        print(missing_uniprots.tolist())
    iuphar_filtered = iuphar_filtered[iuphar_filtered['target_uniprot'].isin(proteins['uniprot'])]
    return iuphar_filtered



from typing import Optional

from click import Context, Argument

from cellphonedb.src.database.manager import DatabaseVersionManager


def choose_database(ctx: Context, argument: Argument, value: str) -> Optional[str]:
    return DatabaseVersionManager.find_database_for(value)


import pandas as pd
from sqlalchemy import or_

from cellphonedb.src.core.database.Repository import Repository
from cellphonedb.src.core.database.sqlalchemy_models.db_model_interaction import Interaction
from cellphonedb.src.core.models.interaction.interaction_helper import expand_interactions_multidatas
from cellphonedb.src.core.utils import filters


class InteractionRepository(Repository):
    name = 'interaction'

    def get_all(self):
        query = self.database_manager.database.session.query(Interaction)
        interactions = pd.read_sql(query.statement, self.database_manager.database.engine)

        return interactions

    def get_interactions_by_multidata_id(self, id):
        """

        :type id: int
        :rtype: pd.DataFrame
        """
        query = self.database_manager.database.session.query(Interaction).filter(
            or_(Interaction.multidata_1_id == int(id), Interaction.multidata_2_id == int(id)))
        result = pd.read_sql(query.statement, self.database_manager.database.engine)

        return result

    def get_interactions_multidata_by_multidata_id(self, id):
        """

        :type id: int
        :rtype: pd.DataFrame
        """

        interactions = self.get_interactions_by_multidata_id(id)
        multidatas_expanded = self.database_manager.get_repository('multidata').get_all_expanded()
        interactions_expanded = expand_interactions_multidatas(interactions, multidatas_expanded)
        return interactions_expanded

    def get_all_expanded(self, include_gene=True, suffixes=('_1', '_2')):
        interactions_query = self.database_manager.database.session.query(Interaction)

        interactions = pd.read_sql(interactions_query.statement, self.database_manager.database.engine)

        multidata_expanded: pd.DataFrame = self.database_manager.get_repository('multidata').get_all_expanded(
            include_gene)

        multidata_expanded = multidata_expanded.astype({'id_multidata': 'int64'})

        interactions = pd.merge(interactions, multidata_expanded, left_on=['multidata_1_id'], right_on=['id_multidata'])
        interactions = pd.merge(interactions, multidata_expanded, left_on=['multidata_2_id'], right_on=['id_multidata'],
                                suffixes=suffixes)

        return interactions

    def add(self, interactions):
        interaction_df = self.blend_dataframes(interactions, ['partner_a', 'partner_b'],
                                               self.database_manager.get_repository('multidata').get_all_name_id(),
                                               'name', 'multidata')

        filters.remove_not_defined_columns(interaction_df,
                                           self.database_manager.get_column_table_names('interaction_table'))

        interaction_df.to_sql(name='interaction_table', if_exists='append', con=self.database_manager.database.engine,
                              index=False, chunksize=50)



from cellphonedb.src.core.core_logger import core_logger
from cellphonedb.src.core.exporters import complex_exporter, interaction_exporter, protein_exporter, gene_exporter, \
    protein_complex_cellphonedb


class ExporterLauncher(object):
    def __init__(self, database_manager):
        self.database_manager = database_manager

    def __getattribute__(self, name):
        method = object.__getattribute__(self, name)
        if hasattr(method, '__call__'):
            core_logger.info('Exporting {}'.format(name))

        return method

    def protein_complex_cellphonedb(self):
        multidatas = self.database_manager.get_repository('multidata').get_all_expanded(include_gene=False)
        interactions = self.database_manager.get_repository('interaction').get_all()
        return protein_complex_cellphonedb.call(multidatas, interactions)


    def complex(self):
        complexes = self.database_manager.get_repository('complex').get_all()
        multidatas = self.database_manager.get_repository('multidata').get_all()
        complex_compositions = self.database_manager.get_repository('complex').get_all_compositions()
        proteins = self.database_manager.get_repository('protein').get_all_expanded()

        return complex_exporter.call(complexes, multidatas, complex_compositions, proteins)

    def interaction(self):
        interactions_expanded = self.database_manager.get_repository('interaction').get_all_expanded(include_gene=False)
        return interaction_exporter.call(interactions_expanded)

    def protein(self):
        proteins_expanded = self.database_manager.get_repository('protein').get_all_expanded()
        return protein_exporter.call(proteins_expanded)

    def gene(self):
        genes_expanded = self.database_manager.get_repository('gene').get_all_expanded()
        output_columns = self.database_manager.get_column_table_names('gene_table') + ['name']
        return gene_exporter.call(genes_expanded, output_columns)



from sqlalchemy import Column, Integer, String, Boolean
from sqlalchemy.orm import relationship

from cellphonedb.src.core.database.sqlalchemy_models import Base


class Multidata(Base):
    __tablename__ = 'multidata_table'
    id_multidata = Column(Integer, nullable=False, primary_key=True)

    name = Column(String, nullable=False, unique=True)

    receptor = Column(Boolean)
    receptor_desc = Column(String)
    other = Column(Boolean)
    other_desc = Column(String)
    secreted_highlight = Column(Boolean)
    secreted_desc = Column(String)
    transmembrane = Column(Boolean)
    secreted = Column(Boolean)
    peripheral = Column(Boolean)
    integrin = Column(Boolean)
    is_complex = Column(Boolean)

    protein = relationship('Protein', backref='protein_table', lazy='subquery')
    complex = relationship('Complex', backref='complex_table', lazy='subquery')



from cellphonedb.src.core.utils import filters


def call(gene_expanded, output_columns):
    filters.remove_not_defined_columns(gene_expanded, output_columns)

    gene_expanded.drop(['id_gene', 'protein_id'], axis=1, inplace=True)

    gene_expanded.rename(index=str, columns={'name': 'uniprot'}, inplace=True)
    return gene_expanded



import pandas as pd
from flask_testing import TestCase

from cellphonedb.src.app.cellphonedb_app import cellphonedb_app
from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.app.flask.flask_app import create_app
from cellphonedb.src.core.database.sqlalchemy_models.db_model_gene import Gene
from cellphonedb.src.core.database.sqlalchemy_models.db_model_multidata import Multidata
from cellphonedb.src.core.database.sqlalchemy_models.db_model_protein import Protein


class TestDatabaseRelationsChecks(TestCase):
    def test_all_protein_have_gen(self):

        expected_protein_without_gene = 0
        protein_query = cellphonedb_app.cellphonedb.database_manager.database.session.query(Protein,
                                                                                            Multidata.name).join(
            Multidata)

        protein_df = pd.read_sql(protein_query.statement,
                                 cellphonedb_app.cellphonedb.database_manager.database.engine)
        protein_ids = protein_df['id_protein'].tolist()

        gene_query = cellphonedb_app.cellphonedb.database_manager.database.session.query(
            Gene.protein_id)
        gene_protein_ids = \
            pd.read_sql(gene_query.statement,
                        cellphonedb_app.cellphonedb.database_manager.database.engine)[
                'protein_id'].tolist()

        protein_without_gene = []
        for protein_id in protein_ids:
            if not protein_id in gene_protein_ids:
                protein_without_gene.append(protein_df[protein_df['id_protein'] == protein_id]['name'].iloc[0])

        if len(protein_without_gene) != expected_protein_without_gene:
            app_logger.warning('There are {} Proteins without gene'.format(len(protein_without_gene)))
            app_logger.warning(protein_without_gene)

            unknowed_proteins_without_gene = []
            for protein in protein_without_gene:
                if not protein in KNOWED_PROTEINS_WITHOUT_GENE:
                    unknowed_proteins_without_gene.append(protein)

            if unknowed_proteins_without_gene:
                app_logger.warning(
                    'There are {} unknowed proteins without gene'.format(len(unknowed_proteins_without_gene)))
                app_logger.warning(pd.Series(unknowed_proteins_without_gene).drop_duplicates().tolist())

        self.assertEqual(expected_protein_without_gene, len(protein_without_gene), 'There are Proteins without Gene.')

    def test_gene_are_not_duplicated(self):
        query = cellphonedb_app.cellphonedb.database_manager.database.session.query(Gene)
        dataframe = pd.read_sql(query.statement,
                                cellphonedb_app.cellphonedb.database_manager.database.engine)

        duplicated_genes = dataframe[dataframe.duplicated(keep=False)]
        if len(duplicated_genes):
            app_logger.warning(duplicated_genes.sort_values('gene_name').to_csv(index=False))

        self.assertEqual(
            len(duplicated_genes), 0,
            'There are %s not not expected duplicated genes in database. '
            'Please check WARNING_duplicated_genes.csv file' % len(duplicated_genes))

    def test_duplicated_gene_ensembl_is_not_in_interaction(self):
        all_genes = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'gene').get_all_expanded()
        all_interactions = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'interaction').get_all()

        genes_duplicated_ensembl = all_genes[all_genes.duplicated('ensembl', keep=False)]

        all_interactions_multidata_ids = all_interactions['multidata_1_id'].tolist() + all_interactions[
            'multidata_2_id'].tolist()

        duplicated_gene_ensembls_in_interactions = genes_duplicated_ensembl[
            genes_duplicated_ensembl['id_multidata'].apply(lambda id: id in all_interactions_multidata_ids)]

        nunknowed_duplicated_ensembl = False
        if not duplicated_gene_ensembls_in_interactions[
            duplicated_gene_ensembls_in_interactions['ensembl'].apply(
                lambda ensembl: ensembl not in KNOWED_DUPLICATED_ENSEMBL.tolist())].empty:
            app_logger.warning('Some duplicated ensembls apears in interactions')
            app_logger.warning(duplicated_gene_ensembls_in_interactions.to_csv(index=False))
            nunknowed_duplicated_ensembl = True

        self.assertFalse(nunknowed_duplicated_ensembl,
                         'Some duplicated ensembl gene apears in interactions')
        self.assertFalse(
            len(KNOWED_DUPLICATED_ENSEMBL.drop_duplicates()) !=
            len(duplicated_gene_ensembls_in_interactions.drop_duplicates('ensembl')),
            'Some duplicated ensembl gene apears in interactions')

    def create_app(self):
        return create_app(raise_non_defined_vars=False, verbose=False)


KNOWED_DUPLICATED_ENSEMBL = pd.Series(['ENSG00000110680'])
KNOWED_PROTEINS_WITHOUT_GENE = []



import pandas as pd

from cellphonedb.src.core.collectors import protein_preprocess_collector, gene_preprocess_collector, \
    complex_preprocess_collector, interaction_preprocess_collector
from cellphonedb.src.core.core_logger import core_logger
from cellphonedb.src.core.database import DatabaseManager


class Collector(object):
    def __init__(self, database_manager: DatabaseManager):
        self.database_manager = database_manager

    def __getattribute__(self, name: str):
        method = object.__getattribute__(self, name)
        if hasattr(method, '__call__'):
            core_logger.info('Collecting {}'.format(name))

        return method

    def protein(self, proteins: pd.DataFrame):
        multidata_columns = self.database_manager.get_column_table_names('multidata_table')
        protein_columns = self.database_manager.get_column_table_names('protein_table')

        proteins_to_add, multidata_to_add = protein_preprocess_collector.call(proteins, multidata_columns,
                                                                              protein_columns)

        self.database_manager.get_repository('protein').add_proteins(proteins_to_add, multidata_to_add)

    def gene(self, genes: pd.DataFrame):
        genes_processed = gene_preprocess_collector.call(genes,
                                                         self.database_manager.get_column_table_names('gene_table'))
        self.database_manager.get_repository('gene').add(genes_processed)

    def complex(self, complexes: pd.DataFrame):
        complexes_processed = complex_preprocess_collector.call(complexes)
        self.database_manager.get_repository('complex').add(complexes_processed)

    def interaction(self, interactions: pd.DataFrame):
        multidatas = self.database_manager.get_repository('multidata').get_all_expanded(include_gene=False)
        interactions_processed = interaction_preprocess_collector.call(interactions, multidatas)
        self.database_manager.get_repository('interaction').add(interactions_processed)

    def all(self, proteins: pd.DataFrame, genes: pd.DataFrame, complexes: pd.DataFrame, interactions: pd.DataFrame):
        self.protein(proteins)
        self.gene(genes)
        self.complex(complexes)
        self.interaction(interactions)



import numpy as np
import pandas as pd
from fbpca import pca
from geosketch import gs

from cellphonedb.src.core.core_logger import core_logger


class Subsampler(object):
    def __init__(self, log: bool, num_pc: int = 100, num_cells: int = None, verbose: bool = None, debug_seed: int = None):
        self.verbose = verbose
        self.log = log
        self.num_pc = num_pc
        self.num_cells = num_cells
        np.random.seed(debug_seed)

    def subsample(self, counts: pd.DataFrame) -> pd.DataFrame:
        input_genes = counts.shape[1]

        if self.num_cells is None:
            self.num_cells = int(input_genes / 3)

        core_logger.info('Subsampling {} to {}'.format(input_genes, self.num_cells))

        counts_t = counts.T

        if self.log:
            pca_input = np.log1p(counts_t)
        else:
            pca_input = counts_t

        try:
            u, s, vt = pca(pca_input.values, k=self.num_pc)
            x_dimred = u[:, :self.num_pc] * s[:self.num_pc]
            sketch_index = gs(x_dimred, self.num_cells, replace=False)
            x_matrix = counts_t.iloc[sketch_index]
        except Exception as e:
            core_logger.warning('Subsampling failed: ignored.')
            if self.verbose:
                core_logger.warning(str(e))
            return counts

        core_logger.info('Done subsampling {} to {}'.format(input_genes, self.num_cells))

        return x_matrix.T



class DatabaseManager:
    def __init__(self, repositories, db=None):
        self._repositories = repositories
        self.database = db

    # TODO: Throw error
    def add_repository(self, repository):
        if not self._repositories:
            self._repositories = {}

        self._repositories[repository.name] = repository

    def get_repository(self, respository_name):
        return self._repositories[respository_name](self)

    def get_column_table_names(self, model_name: str) -> object:
        def get_model():
            for c in self.database.base_model._decl_class_registry.values():
                if hasattr(c, '__tablename__') and c.__tablename__ == model_name:
                    return c

        colum_names = self.database.session.query(get_model()).statement.columns
        colum_names = [p.name for p in colum_names]
        return colum_names



import pandas as pd


def remove_genes_in_file(base_genes: pd.DataFrame, genes_to_remove: pd.DataFrame) -> pd.DataFrame:
    all_genes = pd.merge(base_genes, genes_to_remove, on=['ensembl', 'gene_name', 'hgnc_symbol', 'uniprot'],
                         how='left')

    genes_filtered = all_genes[all_genes['to_keep'] != False]

    if genes_filtered.duplicated('ensembl').any():
        print('WARNING: There are some ensembl duplicated')
        print(
            genes_filtered[genes_filtered.duplicated('ensembl', keep=False)].sort_values('ensembl').to_csv(index=False))

    columns = ['gene_name', 'uniprot', 'hgnc_symbol', 'ensembl']
    return genes_filtered[columns].sort_values(columns, axis=0)



from unittest import TestCase

import pandas as pd

from cellphonedb.src.core.Cellphonedb import data_test_dir
from cellphonedb.src.core.models.cluster_counts import cluster_counts_filter
from cellphonedb.utils import dataframe_functions


class TestFilterClusterCounts(TestCase):
    FIXTURES_SUBPATH = '{}/cluster_counts_model'.format(data_test_dir)

    def test_filter_empty_cluster_counts(self):
        cluster_counts = pd.read_csv('{}/cluster_counts_generic_cluster_counts.csv'.format(self.FIXTURES_SUBPATH))
        expected_result = pd.read_csv(
            '{}/cluster_counts_filter_empty_cluster_results.csv'.format(self.FIXTURES_SUBPATH))

        gene_column_name = 'gene'

        cluster_names = list(cluster_counts.columns.values)
        cluster_names.remove(gene_column_name)

        result = cluster_counts_filter.filter_empty_cluster_counts(
            cluster_counts, cluster_names)

        self.assertTrue(dataframe_functions.dataframes_has_same_data(result, expected_result))



import pandas as pd

from cellphonedb.src.core.database.Repository import Repository
from cellphonedb.src.core.database.sqlalchemy_models.db_model_complex import Complex
from cellphonedb.src.core.database.sqlalchemy_models.db_model_gene import Gene
from cellphonedb.src.core.database.sqlalchemy_models.db_model_multidata import Multidata
from cellphonedb.src.core.database.sqlalchemy_models.db_model_protein import Protein


class MultidataRepository(Repository):
    name = 'multidata'

    def get_all(self):
        query = self.database_manager.database.session.query(Multidata)
        result = pd.read_sql(query.statement, self.database_manager.database.engine)

        return result

    def get_all_expanded(self, include_gene=True):
        protein_multidata_join = Protein.protein_multidata_id == Multidata.id_multidata
        if include_gene:
            gene_protein_join = Gene.protein_id == Protein.id_protein
            query_single = self.database_manager.database.session.query(Gene,
                                                                        Protein,
                                                                        Multidata).join(Protein,
                                                                                        gene_protein_join
                                                                                        ).join(Multidata,
                                                                                               protein_multidata_join)
        else:
            query_single = self.database_manager.database.session.query(Protein, Multidata).join(
                Multidata, protein_multidata_join)

        multidata_simple = pd.read_sql(query_single.statement, self.database_manager.database.engine)

        multidata_complex_join = Multidata.id_multidata == Complex.complex_multidata_id
        query_complex = self.database_manager.database.session.query(Multidata, Complex).join(Complex,
                                                                                              multidata_complex_join)
        multidata_complex = pd.read_sql(query_complex.statement, self.database_manager.database.engine)

        if multidata_complex.empty:
            return multidata_simple

        multidata_expanded = multidata_simple.append(multidata_complex, ignore_index=True, sort=True)

        return multidata_expanded

    def get_all_name_id(self) -> pd.DataFrame:
        query_multidatas = self.database_manager.database.session.query(Multidata.id_multidata, Multidata.name)
        multidatas = pd.read_sql(query_multidatas.statement, self.database_manager.database.session.bind)

        return multidatas

    def get_multidatas_from_string(self, input_string: str) -> pd.DataFrame:
        multidatas = self.get_all_expanded()

        return multidatas[(multidatas['name'] == input_string) |
                          (multidatas['ensembl'] == input_string) |
                          (multidatas['protein_name'] == input_string) |
                          (multidatas['gene_name'] == input_string)]



import pandas as pd


def merge_cellphone_genes(cluster_counts: pd.DataFrame, genes_expanded: pd.DataFrame) -> pd.DataFrame:
    """
    Merges cluster genes with CellPhoneDB values
    """

    multidata_counts = pd.merge(cluster_counts, genes_expanded, left_index=True, right_on='ensembl')

    return multidata_counts


def get_counts_proteins_of_complexes(cluster_counts: pd.DataFrame, multidatas: pd.DataFrame,
                                     complex_composition):
    cluster_clean = cluster_counts.drop(complex_composition.columns.values, errors='ignore', axis=1)

    complex_components_data = pd.merge(multidatas, complex_composition, left_on='id_multidata',
                                       right_on='complex_multidata_id')

    complex_components_data = pd.merge(complex_components_data, cluster_clean, left_on='protein_multidata_id',
                                       right_on='id_multidata', suffixes=['_complex', ''])

    complex_components_data.rename(columns={'name_complex': 'complex_name'}, index=str, inplace=True)

    if complex_components_data.empty:
        return pd.DataFrame()

    result_receptor_complex = complex_components_data.assign(is_complex=True)
    return result_receptor_complex



import pandas as pd
from unittest import TestCase

from cellphonedb.src.core.Cellphonedb import data_test_dir
from cellphonedb.src.core.exceptions.ProcessMetaException import ProcessMetaException
from cellphonedb.src.core.preprocessors import method_preprocessors


class TestMethodPreprocessorsTest(TestCase):

    def test_two_column_both_defined(self):
        self._meta_file_validator('meta_two_columns_both_defined.txt')

    def test_multiple_columns_all_defined(self):
        self._meta_file_validator('meta_multiple_columns_all_defined.txt')

    def test_two_columns_none_defined(self):
        self._meta_file_validator('meta_two_columns_none_defined.txt')

    def test_multiple_columns_only_celltype_defined(self):
        self._meta_file_validator('meta_multiple_columns_celltype_defined.txt')

    def test_incomplete_header(self):
        self._meta_file_validator('meta_multiple_columns_miss_one_header.txt')

    def test_multiple_column_insensitive_case(self):
        self._meta_file_validator('meta_multiple_columns_random_uppercase.txt')

    def test_one_column_fail(self):
        self._meta_file_exception_validator('meta_one_column_fail.txt')

    def test_wrong_format_fail(self):
        self._meta_file_exception_validator('meta_wrong_format_fail.txt')

    def test_multiple_column_miss_multiple_header(self):
        self._meta_file_exception_validator('meta_multiple_columns_miss_multiple_header_fail.txt')

    def _meta_file_exception_validator(self, raw_meta_file):
        raw_meta = pd.read_csv('{}/preprocessors_fixtures/{}'.format(data_test_dir, raw_meta_file), sep='\t')
        with self.assertRaises(ProcessMetaException):
            method_preprocessors.meta_preprocessor(raw_meta)

    def _meta_file_validator(self, raw_meta_file: str) -> None:
        raw_meta = pd.read_csv('{}/preprocessors_fixtures/{}'.format(data_test_dir, raw_meta_file), sep='\t')
        meta = method_preprocessors.meta_preprocessor(raw_meta)

        result_expected = pd.read_csv(
            '{}/preprocessors_fixtures/meta_result_preprocessor.txt'.format(data_test_dir),
            sep='\t', index_col=0)

        self.assertTrue(meta.equals(result_expected))
        self.assertEqual(meta.index.name, 'cell')



import os
from typing import Optional

import pandas as pd

from cellphonedb.src.app.cellphonedb_app import output_test_dir, data_test_dir, cellphonedb_app
from cellphonedb.src.app.flask.flask_app import create_app
from cellphonedb.src.core.utils.subsampler import Subsampler
from cellphonedb.src.local_launchers.local_method_launcher import LocalMethodLauncher
from cellphonedb.src.tests.cellphone_flask_test_case import CellphoneFlaskTestCase
from cellphonedb.utils import dataframe_functions


class TestTerminalMethodlAnalysis(CellphoneFlaskTestCase):
    def create_app(self):
        return create_app(raise_non_defined_vars=False, verbose=False)

    def test_non_statistical_method__data_test__threshold__01__precision_1(self):
        data = 'test'
        project_name = 'test_data'
        threshold = 0.1
        result_precision = 1
        self._method_call(data, project_name, threshold, result_precision)

    def test_non_statistical_method__data_test__threshold__01__precision_1_hgnc(self):
        data = 'test_custom_counts_data'
        project_name = 'test_data'
        threshold = 0.1
        result_precision = 1
        self._method_call(data, project_name, threshold, result_precision, counts_data='hgnc_symbol')

    def test_non_statistical_method__data_test__threshold__01__precision_1_gene_name(self):
        data = 'test_custom_counts_data'
        project_name = 'test_data'
        threshold = 0.1
        result_precision = 1
        self._method_call(data, project_name, threshold, result_precision, counts_data='gene_name')

    def test_non_statistical_method__data_test__threshold__01__precision_3(self):
        data = 'test'
        project_name = 'test_data'
        threshold = 0.1
        result_precision = 3
        self._method_call(data, project_name, threshold, result_precision)

    def test_non_statistical_method_subsampled_data_test__threshold__01__precision_3__num_pc_4__num_cells_4(self):
        data = 'test_subsampled'
        project_name = 'test_data'
        threshold = 0.1
        result_precision = 3
        subsampler = Subsampler(False, 4, 4, debug_seed=0)
        self._method_call(data, project_name, threshold, result_precision, subsampler)

    def _method_call(self, data: str, project_name: str, threshold: float, result_precision: int,
                     subsampler: Optional[Subsampler] = None, counts_data: str = 'ensembl'):
        result_names_as_fixture = False
        if result_names_as_fixture:
            result_deconvoluted_filename, result_means_filename, result_significant_means_filename = self._original_names(
                data, result_precision, threshold)
        else:
            result_means_filename = self._get_result_filename('means', data, threshold, result_precision)
            result_significant_means_filename = self._get_result_filename('significant_means', data, threshold,
                                                                          result_precision)
            result_deconvoluted_filename = self._get_result_filename('deconvoluted', data, threshold, result_precision)

        meta_filename = os.path.realpath('{}/hi_{}_meta.txt'.format(data_test_dir, data))

        if counts_data == 'ensembl':
            counts_file = '{}/hi_{}_counts.txt'.format(data_test_dir, data)
        else:
            counts_file = '{}/hi_{}_counts_{}.txt'.format(data_test_dir, data, counts_data)

        counts_filename = os.path.realpath(counts_file)

        LocalMethodLauncher(cellphonedb_app.cellphonedb).cpdb_analysis_local_method_launcher(meta_filename,
                                                                                             counts_filename,
                                                                                             counts_data,
                                                                                             project_name,
                                                                                             threshold,
                                                                                             output_test_dir,
                                                                                             'txt',
                                                                                             result_means_filename,
                                                                                             result_significant_means_filename,
                                                                                             result_deconvoluted_filename,
                                                                                             result_precision,
                                                                                             subsampler
                                                                                             )

        self._assert_result('means', data, project_name, result_means_filename, threshold, result_precision)
        self._assert_result('significant_means', data, project_name, result_significant_means_filename, threshold,
                            result_precision)
        self._assert_result('deconvoluted', data, project_name, result_deconvoluted_filename, threshold,
                            result_precision)

    def _original_names(self, data, result_precision, threshold):
        str_threshold = ''.join(str(threshold).split('.'))

        result_means_filename = 'analysis__{}_result__data-{}_threshold-{}_precision-{}.txt'.format('means',
                                                                                                    data,
                                                                                                    str_threshold,
                                                                                                    result_precision)
        result_significant_means_filename = 'analysis__{}_result__data-{}_threshold-{}_precision-{}.txt'.format(
            'significant_means',
            data,
            str_threshold,
            result_precision)
        result_deconvoluted_filename = 'analysis__{}_result__data-{}_threshold-{}_precision-{}.txt'.format(
            'deconvoluted',
            data,
            str_threshold,
            result_precision)
        return result_deconvoluted_filename, result_means_filename, result_significant_means_filename

    def _assert_result(self,
                       namefile: str,
                       data: str,
                       project_name: str,
                       result_means_filename: str,
                       threshold: float,
                       result_precision: int,
                       ) -> None:
        str_threshold = ''.join(str(threshold).split('.'))
        test_filename = 'analysis__{}_result__data-{}_threshold-{}_precision-{}.txt'.format(namefile,
                                                                                            data,
                                                                                            str_threshold,
                                                                                            result_precision)
        original_means = pd.read_table('{}/{}'.format(data_test_dir, test_filename))
        result_means = pd.read_table('{}/{}/{}'.format(output_test_dir, project_name, result_means_filename))
        self.assertTrue(dataframe_functions.dataframes_has_same_data(result_means, original_means))
        self.remove_file('{}/{}/{}'.format(output_test_dir, project_name, result_means_filename))

    def _get_result_filename(self, base_name, data: str, threshold: float, precision: int) -> str:
        str_threshold = ''.join(str(threshold).split('.'))
        base_filename = '{}__data-{}_threshold-{}_precision-{}'.format(base_name, data, str_threshold, precision)
        random_filename = self.get_test_filename(base_filename, 'txt')

        return random_filename



from unittest import TestCase

import pandas as pd

from cellphonedb.src.core.Cellphonedb import data_test_dir
from cellphonedb.src.core.models.complex import complex_helper
from cellphonedb.utils import dataframe_functions


class TestHelperComplex(TestCase):
    FIXTURES_SUBPATH = '{}/helper_complex'.format(data_test_dir)

    def test_get_involved_complex_from_protein(self):
        proteins = pd.read_csv('{}/helper_complex_protein.csv'.format(self.FIXTURES_SUBPATH), index_col=0)
        complex_composition = pd.read_csv('{}/helper_complex_complex_composition.csv'.format(self.FIXTURES_SUBPATH))

        result_expected = pd.read_csv('{}/helper_complex_result.csv'.format(self.FIXTURES_SUBPATH), index_col=3)

        result = complex_helper.get_involved_complex_composition_from_protein(proteins, complex_composition)

        self.assertTrue(dataframe_functions.dataframes_has_same_data(result, result_expected))

    def test_get_involved_complex_from_protein_empty_result(self):
        proteins = pd.read_csv('{}/helper_complex_protein.csv'.format(self.FIXTURES_SUBPATH), index_col=0)
        proteins.drop(proteins.index, inplace=True)
        complex_composition = pd.read_csv('{}/helper_complex_complex_composition.csv'.format(self.FIXTURES_SUBPATH))

        result = complex_helper.get_involved_complex_composition_from_protein(proteins, complex_composition)

        self.assertTrue(dataframe_functions.dataframes_has_same_data(result, pd.DataFrame(columns=['complex_multidata_id', 'protein_multidata_id', 'total_protein'])))

    def test_get_involved_complex_composition_from_protein(self):
        proteins = pd.read_csv('{}/helper_complex_protein.csv'.format(self.FIXTURES_SUBPATH), index_col=0)
        complex_composition = pd.read_csv('{}/helper_complex_complex_composition.csv'.format(self.FIXTURES_SUBPATH))

        result_expected = pd.read_csv('{}/helper_complex_result_drop_duplicates.csv'.format(self.FIXTURES_SUBPATH))

        result = complex_helper.get_involved_complex_composition_from_protein(proteins, complex_composition)

        self.assertTrue(dataframe_functions.dataframes_has_same_data(result, result_expected))



import os
import random
import string
import time

from flask_testing import TestCase

from cellphonedb.src.app.cellphonedb_app import cellphonedb_app
from cellphonedb.src.local_launchers.local_collector_launcher import LocalCollectorLauncher
from cellphonedb.utils import utils


class CellphoneFlaskTestCase(TestCase):
    @staticmethod
    def fixtures_dir():
        current_dir = os.path.dirname(os.path.realpath(__file__))
        fixtures_dir = '{}/fixtures'.format(current_dir)

        return fixtures_dir

    @staticmethod
    def reset_db():
        cellphonedb_app.cellphonedb.database_manager.database.drop_everything()
        cellphonedb_app.cellphonedb.database_manager.database.create_all()

    def populate_db(self):
        LocalCollectorLauncher().all('collect_protein.csv', 'collect_gene.csv', 'collect_complex.csv',
                                             'collect_interaction.csv', self.fixtures_dir())

    @staticmethod
    def remove_file(file):
        os.remove(file)

    @staticmethod
    def rand_string(digits=5):
        return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(digits))

    @staticmethod
    def get_test_filename(original_namefile, extension, prefix='TESTING'):
        namefile = '{}_{}_{}_{}.{}'.format(prefix, original_namefile, int(time.time()),
                                           CellphoneFlaskTestCase.rand_string(),
                                           extension)

        return namefile

    def assert_file_not_empty(self, file, message=''):
        if not message:
            message = 'File {} is empty'.format(file)

        read_data = utils.read_data_table_from_file(file)
        self.assertFalse(read_data.empty, message)

    def assert_file_exist(self, path_file, message=''):
        if not message:
            message = 'File {} didnt exist'.format(path_file)

        self.assertTrue(os.path.isfile(path_file), message)



import json
import sys
import traceback

import flask
from flask import request

from cellphonedb.src.app.cellphonedb_app import cellphonedb_app
from cellphonedb.src.api_endpoints.web_api.web_api_endpoint_base import WebApiEndpointBase


class WebEndpointQueryAutocomplete(WebApiEndpointBase):
    def post(self):
        parameters = json.loads(request.get_data(as_text=True))

        partial_element = parameters['partial_element']

        if len(partial_element) < 2:
            return flask.jsonify({'success': True, 'result': []})

        try:
            interactions = cellphonedb_app.cellphonedb.query.autocomplete_launcher(partial_element)
            response = {
                'success': True,
                'result': interactions.to_dict(orient='records')
            }
        except:
            response = {
                'success': False
            }
            print(traceback.print_exc(file=sys.stdout))

        return flask.jsonify(response)



import io
import os
import pickle
from typing import TextIO, Optional

import scipy.io
import pandas as pd
from werkzeug.datastructures import FileStorage

from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.app.cellphonedb_app import output_dir

from cellphonedb.src.exceptions.NotADataFrameException import NotADataFrameException
from cellphonedb.src.exceptions.ReadFileException import ReadFileException
from cellphonedb.src.exceptions.ReadFromPickleException import ReadFromPickleException


def read_data_table_from_file(file: str, index_column_first: bool = False, separator: str = '',
                              dtype=None, na_values=None, compression=None) -> pd.DataFrame:
    if os.path.isdir(file):
        return _read_mtx(file)

    filename, file_extension = os.path.splitext(file)

    if file_extension == '.h5ad':
        return _read_h5ad(file)
    if file_extension == '.h5':
        return _read_h5(file)

    if file_extension == '.pickle':
        try:
            with open(file, 'rb') as f:
                df = pickle.load(f)
                if isinstance(df, pd.DataFrame):
                    return df
                else:
                    raise NotADataFrameException(file)
        except:
            raise ReadFromPickleException(file)

    if not separator:
        separator = _get_separator(file_extension)
    try:
        f = open(file)
    except Exception:
        raise ReadFileException(file)
    else:
        with f:
            return _read_data(f, separator, index_column_first, dtype, na_values, compression)


def read_data_from_content_type(file: FileStorage, index_column_first: bool = False, separator: str = '',
                                dtype=None) -> pd.DataFrame:
    if not separator:
        separator = _get_separator(file.content_type)
    return _read_data(file.stream, separator, index_column_first, dtype)


def read_data_from_s3_object(s3_object: dict, s3_name: str, index_column_first: bool = False, separator: str = '',
                             dtype=None, na_values=None) -> pd.DataFrame:
    filename, file_extension = os.path.splitext(s3_name)
    if not separator:
        separator = _get_separator(file_extension)
        bytestream = io.BytesIO(s3_object['Body'].read())

    return _read_data(bytestream, separator, index_column_first, dtype, na_values)


def write_to_file(df: pd.DataFrame, filename: str, output_path: str, output_format: Optional[str] = None):
    _, file_extension = os.path.splitext(filename)

    if output_format is None:
        if not file_extension:
            default_format = 'txt'
            default_extension = '.{}'.format(default_format)

            separator = _get_separator(default_extension)
            filename = '{}{}'.format(filename, default_extension)
        else:
            separator = _get_separator(file_extension)
    else:
        selected_extension = '.{}'.format(output_format)

        if file_extension != selected_extension:
            separator = _get_separator(selected_extension)
            filename = '{}{}'.format(filename, selected_extension)

            if file_extension:
                app_logger.warning(
                    'Selected extension missmatches output filename ({}, {}): It will be added => {}'.format(
                        selected_extension, file_extension, filename))
        else:
            separator = _get_separator(selected_extension)

    df.to_csv('{}/{}'.format(output_path, filename), sep=separator, index=False)


def _read_mtx(path: str) -> pd.DataFrame:

    mtx_file = path + '/matrix.mtx'
    bc_file = path + '/barcodes.tsv'
    feature_file = path + '/features.tsv'

    df = pd.DataFrame(scipy.io.mmread(mtx_file).toarray())
    bc = [line.strip() for line in open(bc_file)]
    feature = [line.strip() for line in open(feature_file)]

    df.index = feature
    df.columns = bc
    df.index.name = 'Gene'

    return df

def _read_h5ad(path: str) -> pd.DataFrame:
    try:
        from anndata import read_h5ad
    except:
        raise ImportError('Please install anndata. pip install anndata')
    adata = read_h5ad(path)
    df = adata.to_df().T
    return df


def _read_h5(path: str) -> pd.DataFrame:
    df = pd.read_hdf(path)
    return df


def _read_data(file_stream: TextIO, separator: str, index_column_first: bool, dtype=None,
               na_values=None, compression=None) -> pd.DataFrame:
    return pd.read_csv(file_stream, sep=separator, index_col=0 if index_column_first else None, dtype=dtype,
                       na_values=na_values, compression=compression)


def _get_separator(mime_type_or_extension: str) -> str:
    extensions = {
        '.csv': ',',
        '.tsv': '\t',
        '.txt': '\t',
        '.tab': '\t',
        'text/csv': ',',
        'text/tab-separated-values': '\t',
    }
    default_separator = ','

    return extensions.get(mime_type_or_extension.lower(), default_separator)


def set_paths(output_path, project_name):
    if not output_path:
        output_path = output_dir

    if project_name:
        output_path = os.path.realpath(os.path.expanduser('{}/{}'.format(output_path, project_name)))

    os.makedirs(output_path, exist_ok=True)

    if _path_is_not_empty(output_path):
        app_logger.warning(
            'Output directory ({}) exist and is not empty. Result can overwrite old results'.format(output_path))

    return output_path


def _path_is_not_empty(path):
    return bool([f for f in os.listdir(path) if not f.startswith('.')])


import os
from typing import Optional

import pandas as pd

from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.app.cellphonedb_app import output_dir
from cellphonedb.src.core.utils.subsampler import Subsampler
from cellphonedb.utils import utils
from cellphonedb.utils.utils import write_to_file


class LocalMethodLauncher(object):
    def __init__(self, cellphonedb_app):

        self.cellphonedb_app = cellphonedb_app

    def __getattribute__(self, name):
        method = object.__getattribute__(self, name)
        if hasattr(method, '__call__'):
            app_logger.info('Launching Method {}'.format(name))

        return method

    def cpdb_statistical_analysis_local_method_launcher(self, meta_filename: str,
                                                        counts_filename: str,
                                                        counts_data: str,
                                                        project_name: str = '',
                                                        iterations: int = 1000,
                                                        threshold: float = 0.1,
                                                        output_path: str = '',
                                                        output_format: Optional[str] = None,
                                                        means_filename: str = 'means',
                                                        pvalues_filename: str = 'pvalues',
                                                        significant_means_filename: str = 'significant_means',
                                                        deconvoluted_filename='deconvoluted',
                                                        debug_seed: int = -1,
                                                        threads: int = -1,
                                                        result_precision: int = 3,
                                                        pvalue: float = 0.05,
                                                        subsampler: Subsampler = None,
                                                        ) -> None:
        output_path = self._set_paths(output_path, project_name)

        debug_seed = int(debug_seed)
        iterations = int(iterations)
        threads = int(threads)
        threshold = float(threshold)
        result_precision = int(result_precision)

        counts, meta = self._load_meta_counts(counts_filename, meta_filename)

        pvalues_simple, means_simple, significant_means_simple, deconvoluted_simple = \
            self.cellphonedb_app.method.cpdb_statistical_analysis_launcher(
                meta,
                counts,
                counts_data,
                iterations,
                threshold,
                threads,
                debug_seed,
                result_precision,
                pvalue,
                subsampler
            )

        write_to_file(means_simple, means_filename, output_path, output_format)
        write_to_file(pvalues_simple, pvalues_filename, output_path, output_format)
        write_to_file(significant_means_simple, significant_means_filename, output_path, output_format)
        write_to_file(deconvoluted_simple, deconvoluted_filename, output_path, output_format)

    def cpdb_analysis_local_method_launcher(self, meta_filename: str,
                                            counts_filename: str,
                                            counts_data: str,
                                            project_name: str = '',
                                            threshold: float = 0.1,
                                            output_path: str = '',
                                            output_format: Optional[str] = None,
                                            means_filename: str = 'means',
                                            significant_means_filename: str = 'significant_means',
                                            deconvoluted_filename='deconvoluted',
                                            result_precision: int = 3,
                                            subsampler: Subsampler = None,
                                            ) -> None:
        output_path = self._set_paths(output_path, project_name)

        result_precision = int(result_precision)
        threshold = float(threshold)

        counts, meta = self._load_meta_counts(counts_filename, meta_filename)

        means, significant_means, deconvoluted = \
            self.cellphonedb_app.method.cpdb_method_analysis_launcher(meta,
                                                                      counts,
                                                                      counts_data,
                                                                      threshold,
                                                                      result_precision,
                                                                      subsampler)

        write_to_file(means, means_filename, output_path, output_format)
        write_to_file(significant_means, significant_means_filename, output_path, output_format)
        write_to_file(deconvoluted, deconvoluted_filename, output_path, output_format)

    @staticmethod
    def _path_is_empty(path):
        return bool([f for f in os.listdir(path) if not f.startswith('.')])

    @staticmethod
    def _set_paths(output_path, project_name):
        if not output_path:
            output_path = output_dir
        if project_name:
            output_path = os.path.realpath(os.path.expanduser('{}/{}'.format(output_path, project_name)))
        os.makedirs(output_path, exist_ok=True)
        if LocalMethodLauncher._path_is_empty(output_path):
            app_logger.warning(
                'Output directory ({}) exist and is not empty. Result can overwrite old results'.format(output_path))
        return output_path

    @staticmethod
    def _load_meta_counts(counts_filename: str, meta_filename: str) -> (pd.DataFrame, pd.DataFrame):
        """
        :raise ParseMetaException
        """
        meta = utils.read_data_table_from_file(os.path.realpath(meta_filename))
        counts = utils.read_data_table_from_file(os.path.realpath(counts_filename), index_column_first=True)

        return counts, meta



import os

import yaml

from cellphonedb.src.app import app_logger


class AppConfig():
    def __init__(self, environment=None, support=None, load_defaults=None, raise_non_defined_vars=False, verbose=None):

        self._current_dir = os.path.dirname(os.path.realpath(__file__))
        self.config_parameters = self._get_config_parameters(environment, support, load_defaults,
                                                             raise_non_defined_vars)

        self.config = self._load_config(verbose)

        self._set_app_logger_config(self.config['app'])
        self.sqlalchemy_config = {'uri': self._build_sqlalchemy_database_uri(self.config['database']),
                                  'db_core': self.config['database']['db_core'],
                                  'echo': self.config['database']['echo']}

        self.logger_config = self._get_core_logger_config(self.config['app'])
        self.core_config = self._build_core_config()

    def _build_core_config(self):
        core_config = self.config['app']
        core_config['sqlalchemy'] = self.sqlalchemy_config
        core_config['logger'] = self.logger_config

        return core_config

    def get_cellphone_core_config(self):
        return self.core_config

    @staticmethod
    def _get_threads_config(app_config: dict):
        return app_config['threads']

    @staticmethod
    def _set_app_logger_config(app_config: dict):

        level = 'WARNING'

        if app_config['verbose']:
            level = 'INFO'

        if app_config['debug']:
            level = 'DEBUG'

        app_logger.setLevel(level)

    @staticmethod
    def _get_core_logger_config(app_config):
        config = {'level': 'WARNING'}

        if app_config['verbose']:
            config['level'] = 'INFO'

        if app_config['debug']:
            config['level'] = 'DEBUG'

        return config

    @staticmethod
    def _get_config_parameters(environment=None, support=None, load_defaults=None, raise_non_defined_vars=True):
        config_parameters = {}
        config_keys = [{'env_key': 'APP_ENV', 'default': 'core', 'dict_key': 'environment'},
                       {'env_key': 'APP_CONF_SUPPORT', 'default': 'yaml', 'dict_key': 'support'},
                       {'env_key': 'APP_LOAD_DEFAULTS', 'default': 'true', 'dict_key': 'load_defaults'}]

        for config_key in config_keys:
            if locals()[config_key['dict_key']]:
                config_parameters[config_key['dict_key']] = locals()[config_key['dict_key']]
            elif os.environ.get(config_key['env_key']):
                config_parameters[config_key['dict_key']] = os.environ.get(config_key['env_key'])
            else:
                if raise_non_defined_vars:
                    app_logger.app_logger.warning(
                        '{} not defined, setted to {} by default'.format(config_key['env_key'], config_key['default']))
                config_parameters[config_key['dict_key']] = config_key['default']

        return config_parameters

    @staticmethod
    def _load_yaml(yaml_name):
        with open(yaml_name, 'r') as stream:
            try:
                return yaml.load(stream, Loader=yaml.SafeLoader)
            except yaml.YAMLError as exec:
                print(exec)

    def _merge_configs(self, config_base, new_config):
        config = {**config_base, **new_config}
        for key in config_base:
            if isinstance(config_base[key], dict) and key in new_config:
                config[key] = self._merge_configs(config_base[key], new_config[key])

        return config

    def _load_config(self, verbose: bool) -> dict:
        config = {}
        if self.config_parameters['load_defaults']:
            config = self._load_yaml('{}/config/{}.yml'.format(self._current_dir, 'base_config'))
        if self.config_parameters['environment']:
            custom_config = self._load_yaml(
                '{}/config/{}.yml'.format(self._current_dir, self.config_parameters['environment']))
            config = self._merge_configs(config, custom_config)

        if verbose is not None:
            config['app']['verbose'] = verbose

        return config

    @staticmethod
    def _build_sqlalchemy_database_uri(database_config):

        if database_config['adapter'] == 'sqlite' and database_config['db_core']:
            return ''

        if database_config['adapter'] == 'sqlite':
            return '{}:///{}'.format(database_config['adapter'], database_config['path'])

        return '{}://{}:{}@{}:{}/{}'.format(database_config['adapter'], database_config['user'],
                                            database_config['password'], database_config['host'],
                                            database_config['port'], database_config['db_name'])

    def flask_config(self):
        if self.config_parameters['support'] == 'yaml':
            flask_config = self._flask_config_from_yaml(self.config_parameters['environment'],
                                                        self.config_parameters['load_defaults'])
        elif self.config_parameters['support'] == 'environment_vars':
            flask_config = self._flask_config_from_environment_vars(self.config_parameters['load_defaults'])
        else:
            raise Exception('Non valid support defined {yaml,environment_vars}')

        return flask_config

    def _flask_config_from_environment_vars(self, load_defaults):
        flask_config = {}
        if load_defaults:
            flask_config = self._flask_config_from_yaml(None, load_defaults)

        config_keys = ['SECRET_KEY', 'SQLALCHEMY_DATABASE_URI', 'API_PREFIX', 'DEBUG', 'SQLALCHEMY_TRACK_MODIFICATIONS']

        for config_key in config_keys:
            if os.environ.get(config_key):
                flask_config[config_key] = os.environ.get(config_key)
        return flask_config

    def _flask_config_from_yaml(self, environment, load_defaults):

        flask_config = {
            'SECRET_KEY': self.config['flask']['secret_key'],
            'SQLALCHEMY_DATABASE_URI': self._build_sqlalchemy_database_uri(self.config['database']),
            'API_PREFIX': self.config['api']['prefix'],
            'DEBUG': self.config['flask']['secret_key'],
            'SQLALCHEMY_TRACK_MODIFICATIONS': self.config['flask']['sqlalchemy_track_modifications']
        }
        return flask_config



class ThresholdValueException(Exception):
    def __init__(self, threshold_value):
        super(ThresholdValueException, self).__init__(
            'Threshold value ({}) is not valid. Accepted range: 0<=threshold<=1'.format(threshold_value))



import random
from unittest import TestCase

from cellphonedb.src.app.app_config import AppConfig
from cellphonedb.src.core.CellphonedbSqlalchemy import CellphonedbSqlalchemy


class TestAutocompleteQueries(TestCase):

    def setUp(self) -> None:
        self.cellphone = CellphonedbSqlalchemy(AppConfig().get_cellphone_core_config())
        gene_repository = self.cellphone.database_manager.get_repository('gene')
        self.all_genes = gene_repository.get_all_expanded().to_dict(orient='records')

    def test_find_elements_by_gene_name(self):
        self._test_find_elements_by_('gene_name')

    def test_find_elements_by_protein_name(self):
        self._test_find_elements_by_('protein_name')

    def test_find_elements_by_hgnc_symbol(self):
        self._test_find_elements_by_('hgnc_symbol')

    def test_find_elements_by_ensembl(self):
        self._test_find_elements_by_('ensembl')

    def test_find_elements_by_name(self):
        self._test_find_elements_by_('name')

    def _test_find_elements_by_(self, field):
        random_gene = random.choice(self.all_genes)
        whole_input = random_gene[field]

        whole_query_result = self.cellphone.query.autocomplete_launcher(whole_input)

        whole_results = whole_query_result['value'].tolist()
        self.assertIn(whole_input, whole_results)

        partial_input = self._random_substring(whole_input)

        partial_query_result = self.cellphone.query.autocomplete_launcher(partial_input)

        partial_results = partial_query_result['value'].tolist()
        self.assertIn(whole_input, partial_results)

        self.assertGreaterEqual(len(partial_results), len(whole_results))

    def _random_substring(self, whole_input):
        start_index = self._random_position_to_half(whole_input)
        end_index = self._random_position_to_half(whole_input)

        return whole_input[start_index:-end_index if end_index else None]

    @staticmethod
    def _random_position_to_half(string):
        return random.randint(0, int((len(string)) / 2))



import math

import pandas as pd

from cellphonedb.tools.interactions_helper import filter_by_cellphonedb_interactor
from cellphonedb.tools.tools_helper import sort_interactions_partners_alphabetically


def parse_interactions_imex(interactions_base_df, protein_df, gene_df):
    """
    Parses interactions list from IMEx data file.
    Steps:
        1. Get Uniprot values from columns A and B
        2. Get Uniprot values from ensembl in column altA and altB

        3. Remove duplicated interactions:
            - Remove permutated interactions. i.e.:
                    A->B        A->B
                    B->A    =>  A->B
                    A->B        A->B
            - Remove duplicated interactions and merge scores:
                Get maximum score value.
                Intact-score prevails over innatedb score (step 3).

    """
    interactions_base_df.dropna(how='any', subset=['A', 'B'], inplace=True)

    custom_interactions = pd.DataFrame()

    custom_interactions['a_raw_data'] = interactions_base_df['A']
    custom_interactions['b_raw_data'] = interactions_base_df['B']
    custom_interactions['a_raw_ensembl'] = interactions_base_df['altA']
    custom_interactions['b_raw_ensembl'] = interactions_base_df['altB']

    custom_interactions['protein_1'] = interactions_base_df[
        interactions_base_df['A'].apply(lambda value: value.split(':')[0] == 'uniprotkb')]['A'].apply(
        lambda value: value.split(':')[1].split('-')[0])

    custom_interactions['protein_2'] = interactions_base_df[
        interactions_base_df['B'].apply(lambda value: value.split(':')[0] == 'uniprotkb')]['B'].apply(
        lambda value: value.split(':')[1].split('-')[0])

    custom_interactions['annotation_strategy'] = interactions_base_df['provider']

    # Extract ensembl for a_raw_ensembl data. Only if value is not null and has ensembl: prefix
    custom_interactions['ensembl_1'] = custom_interactions.dropna(subset=['a_raw_ensembl'])[
        custom_interactions.dropna(subset=['a_raw_ensembl'])['a_raw_ensembl'].apply(
            lambda value: value.split(':')[0] == 'ensembl')][
        'a_raw_ensembl'].apply(
        lambda value: value.split(':')[1])

    custom_interactions['ensembl_2'] = custom_interactions.dropna(subset=['b_raw_ensembl'])[
        custom_interactions.dropna(subset=['b_raw_ensembl'])['b_raw_ensembl'].apply(
            lambda value: value.split(':')[0] == 'ensembl')][
        'b_raw_ensembl'].apply(
        lambda value: value.split(':')[1])

    custom_interactions = pd.merge(custom_interactions, gene_df, left_on='ensembl_1', right_on='ensembl', how='outer',
                                   indicator='_merge_1')

    custom_interactions.drop(['ensembl'], inplace=True, axis=1)
    custom_interactions = pd.merge(custom_interactions, gene_df, left_on='ensembl_2', right_on='ensembl', how='outer',
                                   indicator='_merge_2', suffixes=['_1', '_2'])

    def get_protein(row, protein_number):
        protein_x = row['protein_%s' % protein_number]
        if isinstance(protein_x, float) and math.isnan(protein_x):
            return row['uniprot_%s' % protein_number]

        return row['protein_%s' % protein_number]

    custom_interactions['protein_1'] = custom_interactions.apply(lambda row: get_protein(row, 1), axis=1)
    custom_interactions['protein_2'] = custom_interactions.apply(lambda row: get_protein(row, 2), axis=1)
    custom_interactions = custom_interactions[['protein_1', 'protein_2', 'annotation_strategy']]

    custom_interactions.dropna(how='any', subset=['protein_1', 'protein_2'], inplace=True)

    # Filtering
    custom_interactions = filter_by_cellphonedb_interactor(protein_df, custom_interactions)

    custom_interactions = sort_interactions_partners_alphabetically(custom_interactions, ('protein_1', 'protein_2'))

    custom_interactions['annotation_strategy'] = custom_interactions.groupby(['protein_1', 'protein_2'])[
        'annotation_strategy'].transform(
        lambda sources: ','.join(sorted(set(sources))))

    custom_interactions_unique = custom_interactions.drop_duplicates(['protein_1', 'protein_2'], keep='first')
    custom_interactions_unique = custom_interactions_unique[['protein_1', 'protein_2', 'annotation_strategy']]

    custom_interactions_unique.rename(index=str, columns={'protein_1': 'uniprot_1', 'protein_2': 'uniprot_2'},
                                      inplace=True)

    return custom_interactions_unique



import os
from io import StringIO

import pandas as pd
import requests
import tqdm

from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.app.cellphonedb_app import data_dir
from cellphonedb.tools.tools_helper import add_to_meta


def call(genes: pd.DataFrame, downloads_path: str, fetch: bool, save_backup: bool = True) -> pd.DataFrame:
    proteins = genes['uniprot'].tolist()

    sources = [
        {
            'name': 'InnateDB',
            'base_url': 'https://psicquic.curated.innatedb.com/webservices/current/search/query/species:human',
            'query_parameters': False,
        },
        {
            'name': 'InnateDB-All',
            'base_url': 'https://psicquic.all.innatedb.com/webservices/current/search/query/species:human',
            'query_parameters': False,
        },
        {
            'name': 'IMEx',
            'base_url': 'http://www.ebi.ac.uk/Tools/webservices/psicquic/imex/webservices/current/search/query/{}',
            'query_parameters': True,
        },
        {
            'name': 'IntAct',
            'base_url': 'http://www.ebi.ac.uk/Tools/webservices/psicquic/intact/webservices/current/search/query/{}',
            'query_parameters': True,
        },
        {
            'name': 'bhf-ucl',
            'base_url': 'http://www.ebi.ac.uk/Tools/webservices/psicquic/bhf-ucl/webservices/current/search/query/{}',
            'query_parameters': True,
        },
        {
            'name': 'MatrixDB',
            'base_url': 'http://matrixdb.univ-lyon1.fr:8080/psicquic/webservices/current/search/query/{}',
            'query_parameters': True,
        },
        {
            'name': 'MINT',
            'base_url': 'http://www.ebi.ac.uk/Tools/webservices/psicquic/mint/webservices/current/search/query/{}',
            'query_parameters': True,
        },
        {
            'name': 'I2D',
            'base_url': 'http://ophid.utoronto.ca/psicquic-ws/webservices/current/search/query/{}',
            'query_parameters': True,
        },
        {
            'name': 'UniProt',
            'base_url': 'http://www.ebi.ac.uk/Tools/webservices/psicquic/uniprot/webservices/current/search/query/{}',
            'query_parameters': True,
        },
        {
            'name': 'MBInfo',
            'base_url': 'http://www.ebi.ac.uk/Tools/webservices/psicquic/mbinfo/webservices/current/search/query/{}',
            'query_parameters': True,
        },
        # 'DIP': 'http://imex.mbi.ucla.edu/xpsq-dip-all/service/rest/current/search/query/{}',
    ]

    significant_columns = ['A', 'B', 'altA', 'altB', 'provider']
    carry = pd.DataFrame(columns=significant_columns)

    for source in tqdm.tqdm(sources, desc='Fetching sources'.ljust(20)):
        carry = carry.append(_get_source(source, proteins, downloads_path, significant_columns, fetch, save_backup),
                             ignore_index=True,
                             sort=False)
    print()

    carry.drop_duplicates(inplace=True)

    return carry


def _get_source(source, proteins, downloads_path, significant_columns, fetch: bool, save_backup: bool = True):
    carry = pd.DataFrame(columns=significant_columns)
    columns_to_save = ['A', 'B', 'altA', 'altB']
    compression = 'xz'
    file_name = '{}_interaction_raw.csv.{}'.format(source['name'], compression)
    download_file_path = os.path.join(downloads_path, file_name)

    def best_path_for(name: str):
        saved_file_path = os.path.join(data_dir, 'sources', name)

        if os.path.exists(download_file_path):
            return download_file_path
        if os.path.exists(saved_file_path):
            return saved_file_path

        app_logger.error('Could not find local source for {}'.format(name))
        exit(1)

    file_path = best_path_for(file_name)

    try:
        if fetch:
            if source['query_parameters']:
                carry = _get_chunked_api_results(carry, columns_to_save, proteins, source)
            else:
                carry = _get_single_api_results(carry, columns_to_save, source)

            carry.drop_duplicates(inplace=True)
            if save_backup:
                carry.to_csv(download_file_path, index=False, compression=compression)
                add_to_meta(file_name, os.path.join(downloads_path, 'meta.json'))

        else:
            tqdm.tqdm.write('Using local version for source {}'.format(source['name']))
            carry = pd.read_csv(file_path, compression=compression)

    except CouldNotFetchFromApiException:
        tqdm.tqdm.write('Could not fetch remote source {}, using available backup'.format(source['name']))
        carry = pd.read_csv(file_path, compression=compression)

    carry['provider'] = source['name']

    return carry


class CouldNotFetchFromApiException(Exception):
    pass


def _get_chunked_api_results(carry, columns_to_save, proteins, source):
    chunk_size = 500
    for idx, chunk in enumerate(tqdm.tqdm(zip(*[iter(proteins)] * chunk_size),
                                          desc=source['name'].ljust(20),
                                          total=int(len(proteins) / chunk_size))):
        url = source['base_url'].format(' or '.join(chunk))
        try:
            response = requests.get(url)

            if response.text:
                s = StringIO(response.text)
                df = pd.read_csv(s, sep='\t', names=columns_to_save, usecols=range(4), na_values='-')
                carry = pd.concat([carry, df], sort=False, axis=0, ignore_index=True)
                carry.drop_duplicates(inplace=True)
            else:
                if response.status_code != 200:
                    raise CouldNotFetchFromApiException()
        except (requests.exceptions.ConnectionError, requests.exceptions.SSLError):
            raise CouldNotFetchFromApiException()

    return carry


def _get_single_api_results(carry, columns_to_save, source):
    url = source['base_url']
    tqdm.tqdm.write('Fetching {}'.format(source['name']))
    try:
        response = requests.get(url)
        if response.text:
            s = StringIO(response.text)
            df = pd.read_csv(s, sep='\t', names=columns_to_save, usecols=range(4), na_values='-')
            carry = pd.concat([carry, df], sort=False, axis=0, ignore_index=True)
            carry.drop_duplicates(inplace=True)
        else:
            if response.status_code != 200:
                raise CouldNotFetchFromApiException()

    except (requests.exceptions.ConnectionError, requests.exceptions.SSLError):
        raise CouldNotFetchFromApiException()

    return carry



from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship

from cellphonedb.src.core.database.sqlalchemy_models import Base


class Protein(Base):
    __tablename__ = 'protein_table'
    id_protein = Column(Integer, nullable=False, primary_key=True)

    protein_name = Column(String)
    tags = Column(String)
    tags_reason = Column(String)
    tags_description = Column(String)
    # pfam = Column(String)

    protein_multidata_id = Column(Integer, ForeignKey('multidata_table.id_multidata'), unique=True, nullable=False)
    gene = relationship('Gene', backref='gene_table', lazy='subquery')



import pandas as pd


def call(complex_compositions: pd.DataFrame, complex_name: str) -> pd.DataFrame:
    complex_elements = complex_compositions[complex_compositions['name_complex'] == complex_name]
    complex_elements = complex_elements[
        ['name_complex', 'name_protein', 'protein_name_protein', 'gene_name_protein']].drop_duplicates()
    return complex_elements
